from __future__ import annotations

import asyncio
from contextlib import contextmanager
from contextvars import ContextVar, Token
from enum import IntEnum, auto
from functools import wraps
from typing import (
    Any,
    AsyncGenerator,
    AsyncIterator,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
)

from restio.cache import ModelCache, QueryCache
from restio.dao import BaseDAO, DAOTask, DAOTaskCallable, check_dao_implemented_method
from restio.event import EventListener
from restio.fields.base import Field, FrozenType, T_co
from restio.graph import (
    DependencyGraph,
    NavigationDirection,
    NavigationType,
    Node,
    Tree,
)
from restio.model import BaseModel, _check_model_type
from restio.query import BaseQuery
from restio.shared import (
    CURRENT_SESSION,
    MODEL_INSTANTIATED_EVENT,
    MODEL_PRE_UPDATE_EVENT,
    MODEL_UPDATE_EVENT,
)
from restio.state import ModelState, ModelStateMachine, Transition

Model_co = TypeVar("Model_co", bound=BaseModel, covariant=True)


class PersistencyStrategy(IntEnum):
    """
    Defines the strategy of error handling during a Session `commit`.

    - INTERRUPT_ON_ERROR (default):
        The Session will be interrupted if any operation made by a DAO throws an
        exception. The operations already performed will be persisted on local cache,
        and the on-going tasks in the other dependency trees will be finalized. The
        commit operation will hang until all running tasks finalize, then a list of all
        DAOTask's will be returned by the `commit`. This is the recommended approach.

    - CONTINUE_ON_ERROR:
        The Session will continue processing all models in the dependency tree that
        have been marked for modification. The framework will persist on local cache
        only the models that have been successfuly synchronized to the remote store. No
        cancellation will be done.
    """

    INTERRUPT_ON_ERROR = auto()
    CONTINUE_ON_ERROR = auto()


class SessionState(IntEnum):
    """
    Stores the current state of the Session. The state is used by the Session
    scope for decision making when particular actions are being executed.

    - STANDBY: The session has been created and is not performing any action.
    - GET: The session is currently acquiring a model from the remote server.
    - ADD: The session is currently adding a new model to the internal cache.
    - COMMIT: The session is performing a commit.
    - ROLLBACK: The session is performing a rollback.
    """

    STANDBY = auto()
    GET = auto()
    ADD = auto()
    COMMIT = auto()
    ROLLBACK = auto()


T = TypeVar("T", bound=Callable[..., Any])


def sessionstate(state: SessionState) -> Callable[[T], T]:
    """
    Decorates the current method to define the state of the session
    during its execution.

    # noqa: DAR301 yield

    :param state: The state to be set to the session while the operation
                  is in place.
    :return: The function decorator.
    """

    @contextmanager
    def _context(session: Session):
        try:
            token = session._state.set(state)
            yield
        finally:
            session._state.reset(token)

    def deco(func: T) -> T:
        if not asyncio.iscoroutinefunction(func):

            @wraps(func)
            def wrapper(self, *args, **kwargs):
                with _context(self):
                    return func(self, *args, **kwargs)

        else:

            @wraps(func)
            async def wrapper(self, *args, **kwargs):
                with _context(self):
                    return await func(self, *args, **kwargs)

        return wrapper

    return deco


class SessionException(BaseException):
    """
    Raised when at least one Exception is raised by a DAOTask during a commit.
    """

    exception_tasks: List[Tuple[DAOTask, BaseException]]
    sucessful_tasks: List[DAOTask]

    def __init__(
        self,
        _exception_tasks: List[Tuple[DAOTask, BaseException]],
        _successful_tasks: List[DAOTask],
    ) -> None:
        super().__init__("One or more DAOTasks failed during the commit.")

        self.exception_tasks = _exception_tasks
        self.sucessful_tasks = _successful_tasks


class Session:
    """
    Manages a local session scope for interfacing with a remote REST API server.

    The Session will manage an internal cache where it stores the models retrieved
    from or persisted to the remote REST API. Models are uniquely identified by their
    internal id and their primary keys. Trying to retrieve the same model twice within
    the same scope will make the Session return the cached models instead of querying
    the remote server again.

    When all models have been modified, created or deleted, the Session can be
    persisted on the remote server with `commit`.

    The Session should know how to manipulate each model type by registering the
    respective DAOs with `register_dao`. When asked to `get`, `add` or `remove` a
    model, the Session will look up for the DAO associated with the provided model
    type. If the DAO raises an exeption during `commit`, then the Session will handle
    the erros based on the predefined PersistenceStrategy provided during
    instantiation.

    `get` operations will be executed on spot, while `add`, `remove` and `_update` will
    be scheduled to run during `commit`. The Session will decide in which order to
    manipulate each model based on dependency trees generated in runtime. The Session
    will try to parallelize the calls to the remote server as much as possible to
    guarantee data consistency. The dependency trees vary with the current internal
    state of each model instance as soon as `commit` is called. The new state of each
    model is managed by the Session itself during the commit.

    When called by a Session, DAO instances are allowed to modify the models locally
    during `get`, `add` or `update` without causing the models' states to change, and
    without being affected by the defined frozen attribute of each field.

    After a `commit` is done, all models that have been persisted on the remote server
    successfully will be persisted on the local cache, even if errors occur in other
    parallel tasks.
    """

    _model_cache: ModelCache
    _query_cache: QueryCache
    _daos: Dict[Type[BaseModel], BaseDAO]
    _strategy: PersistencyStrategy
    _state: ContextVar[SessionState]
    _model_lock: Dict[str, asyncio.Lock]
    _session_token: Token[Optional[Session]]
    _listener: EventListener

    def __init__(
        self, strategy: PersistencyStrategy = PersistencyStrategy.INTERRUPT_ON_ERROR
    ):
        self._model_cache = ModelCache()
        self._query_cache = QueryCache()
        self._daos = {}
        self._strategy = strategy
        self._state = ContextVar("SessionStateContext", default=SessionState.STANDBY)
        self._model_lock = {}
        self._listener = EventListener()

        self._listener.subscribe(MODEL_INSTANTIATED_EVENT, self._add_from_context)

    def reset(self):
        """
        Resets the internal cache. All references to models are lost.
        """
        for model in self._model_cache.get_all_models():
            model._state = ModelState.DISCARDED
            self.unregister_model(model)

        self._model_cache.reset()
        self._query_cache.reset()
        self._model_lock = {}

    @property
    def state(self) -> SessionState:
        """
        Returns the state in which the Session is within the current context.

        :return: The SessionState set for the context.
        """
        return self._state.get()

    def register_dao(self, dao: BaseDAO[BaseModel]):
        """
        Registers a DAO instance to the session and maps its model type in the local
        dictionary.

        :param dao: The DAO instance.
        """
        model_type = dao._model_type

        if model_type not in self._daos:
            self._daos[model_type] = dao
            dao.session = self

    def get_dao(self, model_type: Type[Model_co]) -> BaseDAO[Model_co]:
        """
        Returns the DAO associated with `model_type`.

        :param model_type: The BaseModel subclass.
        :raises NotImplementedError: If no DAO has been associated to the provided
                                     `model_type`.
        :return: The DAO instance.
        """
        dao = self._daos.get(model_type)
        if dao is None:
            raise NotImplementedError(
                f"Model type `{model_type.__name__}` does not contain a DAO"
                " registered."
            )

        return dao

    @sessionstate(SessionState.GET)
    async def get(self, model_type: Type[Model_co], **keys) -> Model_co:
        """
        Tries retrieving the model of type `model_type` and primary keys `keys` from
        the local cache. If not found, then calls the `get` method of the DAO
        associated to the `model_type`.

        :param model_type: The BaseModel subclass representing the model.
        :param keys: The primary key values that represents the model.
        :raises ValueError: When a primary key value is missing.
        :raises RuntimeError: When no model is found with the given primary keys.
        :return: The model instance.
        """
        pks_dict: Dict[str, Any] = {}
        for pk in model_type._meta.primary_keys.values():
            if pk.name not in keys:
                raise ValueError(
                    f"Expected value for primary key `{pk.name}` of model type "
                    f" {model_type.__class__.__name__} not provided."
                )
            pks_dict[pk.name] = keys[pk.name]

        model_lock = self._get_model_lock(model_type, **pks_dict)

        async with model_lock:
            pks: Tuple[Any, ...] = tuple(pks_dict.values())
            model: Optional[Model_co] = self._model_cache.get_by_primary_key(
                (model_type, pks)
            )
            if model:
                return model

            dao: BaseDAO = self.get_dao(model_type)
            check_dao_implemented_method(dao.get)

            model = await dao.get(**pks_dict)
            if model:
                model._state = ModelStateMachine.transition(
                    Transition.GET_OBJECT, model._state
                )
                self.register_model(model)
                return model

        keys_str = ", ".join("{}={}".format(k, str(v)) for k, v in pks_dict.items())
        raise RuntimeError(
            f"No model of type `{model_type.__class__.__name__}`` with primary keys "
            f"({keys_str}) was found"
        )

    def _get_model_lock(self, model_type: Type[BaseModel], **keys) -> asyncio.Lock:
        keys_str = ", ".join("{}={}".format(k, str(v)) for k, v in keys.items())
        lock_hash = f"{model_type.__name__}/{str(keys_str)}"
        return self._model_lock.setdefault(lock_hash, asyncio.Lock())

    @sessionstate(SessionState.ADD)
    def add(self, model: BaseModel):
        """
        Adds the new `model` to the local cache and schedules for adding in the remote
        server during `commit`.

        :param model: The model instance.
        :raises RuntimeError: If the model has been already registered in cache.
        """
        _check_model_type(model)

        dao = self.get_dao(type(model))
        check_dao_implemented_method(dao.add)

        if self._model_cache.is_registered_by_id(model):
            raise RuntimeError(
                f"Model of id `{model._internal_id}` is already registered in internal"
                " cache."
            )

        obj_type, obj_keys = key_hash = self._model_cache._get_type_key_hash(model)
        if obj_keys:
            if self._model_cache.has_keys(key_hash):
                raise RuntimeError(
                    f"Model with type `{obj_type.__class__.__name__}` and keys"
                    " `{obj_keys}` is already registered in internal cache."
                )

        for field in model._meta.fields.values():
            self._check_frozen_field(model, field)

        model._state = ModelStateMachine.transition(Transition.ADD_OBJECT, model._state)
        self.register_model(model, force=False)

    def _add_from_context(self, model: BaseModel):
        if self.state == SessionState.STANDBY:
            self.add(model)

    def remove(self, model: BaseModel) -> bool:
        """
        Schedules model for removal on the remote server during `commit`.

        :param model: The model instance.
        :return: True if model is scheduled for removal, False otherwise.
        """
        dao = self.get_dao(type(model))
        check_dao_implemented_method(dao.remove)

        self._assert_cache_internal_id(model)
        model._state = ModelStateMachine.transition(
            Transition.REMOVE_OBJECT, model._state
        )

        return model._state == ModelState.DELETED

    def _assert_cache_internal_id(self, model: BaseModel):
        _check_model_type(model)

        if not self._model_cache.is_registered_by_id(model):
            raise RuntimeError(
                f"Model with internal id {model._internal_id} not found on cache."
            )

    def _pre_update(self, model: Model_co, field: Field[T_co], value: T_co):
        self._check_frozen_field(model, field)

    def _check_frozen_field(self, model: BaseModel, field: Field):
        # During COMMIT, ROLLBACK and GET, all changes to models
        # should be allowed, regardless of the field being or not frozen
        if self.state in (
            SessionState.COMMIT,
            SessionState.ROLLBACK,
            SessionState.GET,
        ):
            return

        if field.frozen == FrozenType.NEVER:
            return

        frozen_state: FrozenType
        if (
            model._state == ModelState.CLEAN
            or model._state == ModelState.DIRTY
            or model._state == ModelState.DELETED
        ):
            frozen_state = FrozenType.UPDATE
        elif model._state == ModelState.NEW:
            frozen_state = FrozenType.CREATE
        elif self.state == SessionState.ADD:
            frozen_state = FrozenType.CREATE
        else:  # DISCARDED
            return

        if field.frozen & frozen_state:  # type: ignore
            if self.state == SessionState.ADD:
                # this will kick in if a model is being registered to be added,
                # but a frozen field has a value different from its default
                field_value = model.fields[field.name]
                if field_value != field.default:
                    raise ValueError(
                        f"Field {field._field_name(model)} is frozen and its value"
                        " cannot be different from its default."
                    )
            else:
                # all other update use cases
                raise ValueError(
                    f"Field {field._field_name(model)} is frozen and cannot be"
                    " modified."
                )

    def _update(self, model: BaseModel, field: Field[T_co], value: T_co) -> bool:
        # check if any value being passed is actually a model that is already
        # registered in the cache, otherwise complains
        if field.depends_on:
            if isinstance(value, Iterable):
                for v in value:
                    self._check_model_is_valid(v)
            elif isinstance(value, BaseModel):
                self._check_model_is_valid(value)

        # During ROLLBACK and GET, all changes to models should be permanent, therefore
        # we persist any change and avoid changing the state of the models
        if self.state in (SessionState.ROLLBACK, SessionState.GET):
            model._persist()
            return False

        model._update_persistent_values(field, value)

        updated = False

        new_state = ModelStateMachine.transition(Transition.UPDATE_OBJECT, model._state)
        if new_state in (ModelState.DIRTY, ModelState.NEW):
            updated = bool(model._persistent_values)
            if not updated and new_state == ModelState.DIRTY:
                new_state = ModelStateMachine.transition(
                    Transition.CLEAN_OBJECT, model._state
                )
            elif new_state == ModelState.NEW:
                model._persist()

        if new_state == ModelState.DIRTY:
            dao = self.get_dao(type(model))
            check_dao_implemented_method(dao.update)

        model._state = new_state
        return updated

    @sessionstate(SessionState.GET)
    async def query(
        self, query: BaseQuery[Model_co], force: bool = False
    ) -> Tuple[Model_co, ...]:
        """
        Runs custom query `query` and registers results in the local cache.

        When new queries are run for the first time, the results are persisted into the
        query and model caches. Models that already exist in cache (by checking their
        primary keys) are not registered again (even if `force=True`), to preserve the
        information already stored and avoid discrepancies.

        The returning values from the query will contain only the models that are
        registered in the cache. Discarded models are replaced with their cached
        version.

        This method will always return an ordered tuple containing all the values as
        they were registered in cache, even when the return type of the query is
        another iterable other than tuple.

        :param query: The query instance to be executed.
        :param force: Forces running the query even if it has been already cached.
        :raises TypeError: When the provided `query` is not a BaseQuery.
        :return: The list of models retrieved by the query.
        """
        if not isinstance(query, BaseQuery):
            raise TypeError("The provided `obj` should be an instance of BaseQuery")

        if not force:
            cached_results = self._query_cache.get(query)
            if cached_results is not None:
                return cached_results

        results = await query(self)
        await self.register_query(query, results, force=False)

        # Registering the cache above without forcing the values
        # into the model cache might cause double instances of the
        # same model to exist (one from the incoming query, and one
        # previously registered by another query or a get). Therefore,
        # results from the query should updated to return the
        # combination between new results from the current query and
        # old results coming from the cache. As this is already handled
        # by register_query, all we need to do is to retrieve the final
        # values stored for the query in cache
        return self._query_cache.get(query)  # type: ignore

    def register_model(self, model: BaseModel, force: bool = False) -> bool:
        """
        Registers the `model` in the internal cache. If the model is already registered
        and `force` is False, then the operation is skipped.

        :param model: The model instance to be registered.
        :param force: Forces the operation if True, skips otherwise. Defaults to False.
        :raises RuntimeError: When at least one child of the provided model is not yet
                              registered in the cache.
        :return: True if the model has been registered, False if it has been skipped.
        """
        if model._state == ModelState.DISCARDED:
            raise RuntimeError(
                f"Model of id `{model._internal_id}` has been discarded and cannot be"
                " registered again to a Session."
            )

        self._check_models_children({model}, recursive=True)

        registered = self._model_cache.register(model, force)
        if registered:
            model._state = ModelStateMachine.transition(
                Transition.REGISTER_OBJECT, model._state
            )
            self._subscribe_update(model)

        return registered

    def unregister_model(self, model: BaseModel):
        """
        Unregisters model from the internal cache.

        :param model: The model instance to be unregistered.
        """
        try:
            self._model_cache.unregister(model)
        except ValueError:
            pass
        finally:
            self._unsubscribe_update(model)

    def _subscribe_update(self, model: BaseModel):
        model._listener.subscribe(MODEL_PRE_UPDATE_EVENT, self._pre_update)
        model._listener.subscribe(MODEL_UPDATE_EVENT, self._update)

    def _unsubscribe_update(self, model: BaseModel):
        model._listener.unsubscribe(MODEL_PRE_UPDATE_EVENT, self._pre_update)
        model._listener.unsubscribe(MODEL_UPDATE_EVENT, self._update)

    async def register_query(
        self,
        query: BaseQuery[Model_co],
        models: Iterable[Model_co],
        force: bool = False,
    ):
        """
        Registers the `query` and its `models` in the internal cache. If any of the
        models is already registered and `force` is False, then they are not persisted
        into the model cache again.

        All query results are stored as tuples, to preserve the order in an immutable
        structure.

        :param query: The query instance to be registered.
        :param models: The iterable containing the models to be registered as a result
                       of the `query`. The order of iteration is preserved in the cache
                       as a tuple.
        :param force: Forces the operation if True, skips otherwise. Defaults to False.
        """
        registered_models: Tuple[Model_co, ...] = tuple()

        for model in models:
            if self._model_cache.has_model_with_keys(model):
                registered_model = await self.get(type(model), **model.primary_keys)
            else:
                model._state = ModelStateMachine.transition(
                    Transition.GET_OBJECT, model._state
                )
                self.register_model(model, force)
                registered_model = model

            if registered_model:
                registered_models += (registered_model,)

        self._query_cache.register(query, registered_models)

    @sessionstate(SessionState.COMMIT)
    async def commit(self, raise_for_status: bool = True) -> List[DAOTask]:
        """
        Persists all models on the remote server. Models that have been successfully
        submited are also persisted on the local cache by the end of the operation.

        :param raise_for_status: The commit will raise a SessionException if at
                                 least one task executed in the commit raises an
                                 exception.
        :return: A list containing all DAOTask's performed by the commit, in the order
                 in which the operations have been finalized.
        """
        cached_models = self._model_cache.get_all_models()

        self._check_models_consistency(cached_models)

        models_to_add = self._get_models_to_add(cached_models)
        models_to_update = self._get_models_to_update(cached_models)
        models_to_remove = self._get_models_to_remove(cached_models)

        all_models: Set[BaseModel] = models_to_add.union(models_to_update).union(
            models_to_remove
        )

        # add, update and remove get one graph each, as each level needs to be
        # completely finished in order for the next to proceed
        graphs = [
            DependencyGraph.generate_from_objects(m)
            for m in (models_to_add, models_to_update, models_to_remove)
        ]

        directions: List[NavigationDirection] = [
            NavigationType.LEAVES_TO_ROOTS,
            NavigationType.LEAVES_TO_ROOTS,
            NavigationType.ROOTS_TO_LEAVES,
        ]

        results: List[DAOTask] = []
        try:
            async for task in self._process_all_graphs(graphs, directions):
                results.append(task)
        finally:
            # clean up models that were discarded
            discarded_models: Set[BaseModel] = set(
                m for m in all_models if m._state == ModelState.DISCARDED
            )
            for model in discarded_models:
                self.unregister_model(model)
            self.update_cache()

        if raise_for_status:
            self.raise_for_status(results)

        return results

    def _check_models_consistency(self, models: Set[BaseModel]):
        for model in models:
            self._check_model_is_valid(model)

        self._check_models_children(models, recursive=False)

        self._check_deleted_models(models)

    def _check_models_children(self, models: Set[BaseModel], recursive: bool):
        for model in models:
            for child in model.get_children(recursive=recursive):
                self._check_model_is_valid(child)

    def _check_model_is_valid(self, model: BaseModel):
        status_valid = model._state not in (ModelState.DISCARDED, ModelState.UNBOUND,)

        if not status_valid:
            raise ValueError(
                f"Model of id `{model._internal_id}` ({model.__class__.__name__}) is"
                f" not valid because its state is either DISCARDED or UNBOUND."
            )

        if not self._model_cache.is_registered_by_id(model):
            raise ValueError(
                f"Model of id `{model._internal_id}` ({model.__class__.__name__}) is"
                f" not valid because it is not registered to the cache."
            )

    @staticmethod
    def _check_deleted_models(models: Set[BaseModel]):
        nodes = DependencyGraph._get_connected_nodes(models)
        deleted_nodes = (n for n in nodes if n.node_object._state == ModelState.DELETED)
        for node in deleted_nodes:
            for parent_node in node.get_parents(recursive=True):
                parent_model: BaseModel = parent_node.node_object
                if parent_model._state not in (
                    ModelState.DELETED,
                    ModelState.DISCARDED,
                ):
                    raise RuntimeError(
                        "Inconsistent tree. Models that are referred by "
                        "other models cannot be deleted."
                    )

    async def _process_all_graphs(
        self, graphs: List[DependencyGraph], directions: List[NavigationDirection]
    ):
        cancel = False

        for graph, direction in zip(graphs, directions):
            # this will avoid moving to the next graph if an error
            # occurs and the cancellation is enabled
            if cancel:
                break

            # processes next level, one at a time
            async for dao_task in self._process_all_trees(graph, direction):
                model = None
                try:
                    # we await here to check if any exceptions were raised
                    # and also to collect the node that was processed
                    await dao_task
                    model = dao_task.model
                except Exception:
                    # INTERRUPT_ON_ERROR will only allow trees on the same
                    # graph to finalize their on going processes - therefore
                    # further levels need to be skipped
                    if self._strategy == PersistencyStrategy.INTERRUPT_ON_ERROR:
                        cancel = True
                    # CONTINUE_ON_ERROR will keep going - this statement is
                    # superfluous, but was placed to make the operation explicit
                    elif self._strategy == PersistencyStrategy.CONTINUE_ON_ERROR:
                        continue
                finally:
                    if model:
                        model._persist()
                        model._state = ModelStateMachine.transition(
                            Transition.PERSIST_OBJECT, model._state
                        )
                    yield dao_task

    async def _process_all_trees(
        self, graph: DependencyGraph, direction: NavigationDirection
    ) -> AsyncIterator[DAOTask]:

        all_results: List[DAOTask] = []

        # this will iterate over each tree separately
        async def iterate_tree(tree, direction):
            tree_results = []
            async for dao_task in self._process_tree(tree, direction):
                tree_results.append(dao_task)
            return tree_results

        # the following block will guarantee that the trees are triggered in parallel
        # and operations run as fast as possible
        tree_iterate_tasks = [
            asyncio.create_task(iterate_tree(t, direction)) for t in graph.trees
        ]
        for iterate_task in asyncio.as_completed(tree_iterate_tasks):
            all_results.extend(await iterate_task)

        # with the parallelization, we lost control of the order in which the models
        # were processe between the trees, so we need to reorder and yield back
        for x in sorted(all_results):
            yield x

    async def _process_tree(
        self, tree: Tree[Model_co], direction: NavigationDirection
    ) -> AsyncGenerator[DAOTask, None]:
        navigate_queue: asyncio.Queue[Node[Model_co]] = asyncio.Queue()
        finished_queue: asyncio.Queue[DAOTask[Model_co]] = asyncio.Queue()
        tasks: List[asyncio.Task] = []
        cancel = False

        # collects nodes to be processed
        nodes = tree.get_nodes()
        nodes_callables: Dict[Node, DAOTask] = self._get_dao_tasks(nodes)

        # creates a task to handle the node and propagate
        # the values across the queues
        async def node_task(node) -> None:
            nonlocal cancel
            dao_task: DAOTask = nodes_callables[node]

            # we don't care about the value of the task here, we only
            # try and except to catch errors and interrupt the execution
            # if necessary
            try:
                await dao_task
            except Exception:
                # in this case, we flag to interrupt scheduling new tasks
                if self._strategy == PersistencyStrategy.INTERRUPT_ON_ERROR:
                    cancel = True

            # informs tree.navigate that a new node has been processed
            await navigate_queue.put(node)
            # queues up a new value to be yielded
            await finished_queue.put(dao_task)

        async for node in tree.navigate(nodes, direction, navigate_queue):
            # yields the processed values before scheduling
            # a new task
            while not finished_queue.empty():
                yield await finished_queue.get()

            # stops scheduling new values if the task is canceled
            if cancel:
                break

            # schedules a new DAO task
            task = asyncio.create_task(node_task(node))
            tasks.append(task)

        # processes the remaining items after the generator has been
        # finalized
        while not finished_queue.empty():
            yield await finished_queue.get()

    def _get_dao_tasks(
        self, nodes: Set[Node[Model_co]]
    ) -> Dict[Node[Model_co], DAOTask[Model_co]]:
        callables = {}

        for node in nodes:
            model = node.node_object
            dao = self.get_dao(type(model))

            model_callable: DAOTaskCallable

            if model._state == ModelState.NEW:
                model_callable = dao.add
            elif model._state == ModelState.DIRTY:
                model_callable = dao.update
            elif model._state == ModelState.DELETED:
                model_callable = dao.remove
            else:
                continue

            callables[node] = DAOTask(node, model_callable)

        return callables

    def update_cache(self):
        """
        Inspects all processed models during a commit an forces registering again.
        This is mainly used to guarantee that models with changed primary keys have
        been re-registered in the cache with the correct value.
        """
        models = self._get_clean_models()
        for model in models:
            self._model_cache.register(model, force=True)

    @staticmethod
    def raise_for_status(tasks: Iterable[DAOTask]):
        """
        Raises a SessionException if at least one resulting DAOTask from `tasks`
        returned by Session.commit() contains an underlying exception raised.

        :param tasks: The list of DAOTasks incoming from Session.commit()
        :raises SessionException: When at least one task has raised an exception.
        """
        exception_tasks = []
        successful_tasks = []

        for task in tasks:
            if task.task.exception():
                exception_tasks.append((task, task.task.exception()))
            else:
                successful_tasks.append(task)

        if exception_tasks:
            raise SessionException(exception_tasks, successful_tasks)

    @sessionstate(SessionState.ROLLBACK)
    def rollback(self):
        """
        Discards all changes from the local cache. New models are discarded and local
        changes to models retrieved from the remote are ignored.
        """
        for model in self._model_cache.get_all_models():
            model._state = ModelStateMachine.transition(
                Transition.ROLLBACK_OBJECT, model._state
            )
            if model._state == ModelState.DISCARDED:
                self.unregister_model(model)
            else:
                model._rollback()

    def _get_models_to_add(
        self, models: Optional[Set[BaseModel]] = None
    ) -> Set[BaseModel]:
        return self._get_models_by_state((ModelState.NEW,), models)

    def _get_models_to_update(
        self, models: Optional[Set[BaseModel]] = None
    ) -> Set[BaseModel]:
        return self._get_models_by_state((ModelState.DIRTY,), models)

    def _get_models_to_remove(
        self, models: Optional[Set[BaseModel]] = None
    ) -> Set[BaseModel]:
        return self._get_models_by_state((ModelState.DELETED,), models)

    def _get_clean_models(
        self, models: Optional[Set[BaseModel]] = None
    ) -> Set[BaseModel]:
        return self._get_models_by_state((ModelState.CLEAN,), models)

    def _get_models_by_state(
        self, state: Tuple[ModelState, ...], models: Optional[Set[BaseModel]] = None
    ):
        models = self._model_cache.get_all_models() if not models else models
        return set(model for model in models if model._state in state)

    async def __aenter__(self):
        self._session_token = CURRENT_SESSION.set(self)
        return self

    async def __aexit__(self, type, value, traceback):
        # exception raised in the context, so we just need to propagate
        # the exception back after rolling back the session
        if value:
            self.rollback()
            CURRENT_SESSION.reset(self._session_token)
            return False

        # processing of the context went well, so we proceed and commit the
        # exception that will be rolled back in case of errors
        try:
            await self.commit(raise_for_status=True)
        except SessionException:
            self.rollback()
            raise
        finally:
            CURRENT_SESSION.reset(self._session_token)

        return True



from contextvars import ContextVar
from typing import TYPE_CHECKING, Dict, Optional, Type

if TYPE_CHECKING:
    from restio.session import Session

CURRENT_SESSION: ContextVar[Optional["Session"]] = ContextVar(
    "current_session", default=None
)


# Stores the names for a model class type
# this object stays here and is globaly declared so that it can be
# accessed by both BaseModelMeta and Field instances
# In case we need to do any further business with this object,
# then implement a Singleton class to handle the object in a more
# structured way
MODEL_TYPE_REGISTRY: Dict[str, Type] = {}


MODEL_INSTANTIATED_EVENT = "__init__"
MODEL_PRE_UPDATE_EVENT = "__pre_update__"
MODEL_UPDATE_EVENT = "__updated__"



from typing import Any, Dict, Optional, Set, Tuple, Type, TypeVar

from restio.model import BaseModel, _check_model_type
from restio.query import BaseQuery

Model_co = TypeVar("Model_co", bound=BaseModel, covariant=True)

KeyCacheKey = Tuple[Type[Model_co], Tuple[Any, ...]]


class ModelCache:
    """
    Stores BaseModel objects in dictionary with (type, hash) indexes.
    """

    _id_cache: Set[BaseModel]
    _key_cache: Dict[KeyCacheKey, BaseModel]

    def __init__(self):
        self.reset()

    def reset(self):
        """
        Resets the internal cache.
        """
        self._id_cache = set()
        self._key_cache = {}

    def register(self, obj: BaseModel, force: bool = False) -> bool:
        """
        Registers a model into the internal cache.

        :param obj: The model to be registered.
        :param force: Forces the registration of the model again if it is already in
                      cache. Defaults to False.
        :return: True if the model has been registered. False otherwise.
        """
        in_id_cache: bool = self.is_registered_by_id(obj)
        model_key: Optional[KeyCacheKey]

        if force:
            model_key = self._search_key_for_model(obj, deep_search=True)
            self._remove_from_cache(obj, model_key)
            in_id_cache = False

        model_key = self._get_type_key_hash(obj)

        # inserts the model if not in cache at this point
        if not in_id_cache and (model_key not in self._key_cache or force):
            self._id_cache.add(obj)
            if not self._has_empty_pk(model_key):
                self._key_cache[model_key] = obj
            return True

        return False

    def unregister(self, obj: BaseModel):
        """
        Unregisters a model from cache.

        :param obj: The model to be unregistered.
        :raises ValueError: When model is not found in cache.
        """

        if not self.is_registered_by_id(obj):
            raise ValueError(
                f"Provided model `{obj._internal_id}` is not registered to the cache."
            )

        cached_model_key = self._search_key_for_model(obj, deep_search=True)

        # now remove the cached models
        self._remove_from_cache(obj, cached_model_key)

    def has_model(self, model: BaseModel) -> bool:
        """
        Indicates if model exists in cache.

        :param model: The model instance.
        :return: True if model exists in cache, False otherwise.
        """
        return self.is_registered_by_id(model)

    def has_model_with_keys(self, model: BaseModel):
        """
        Indicates if a model is registered under the keys of the provided `model`.

        :param model: The model instance.
        :return: True if any model with the keys of `model` is found in the model
                 cache, False otherwise.
        """
        key_hash = self._get_type_key_hash(model)
        return self.has_keys(key_hash)

    def has_keys(self, key_hash: KeyCacheKey) -> bool:
        return not self._has_empty_pk(key_hash) and key_hash in self._key_cache

    def get_by_primary_key(self, key_hash: KeyCacheKey) -> Optional[BaseModel]:
        """
        Finds model in cache by its primary key.

        :param key_hash: The hash representing the primary key in cache.
        :return: If found in cache, returns the model instance. Returns None otherwise.
        """
        return self._key_cache.get(key_hash, None)

    def is_registered_by_id(self, obj: BaseModel) -> bool:
        """
        Indicates if the model `obj` is registered in the Id cache.

        :param obj: The model.
        :return: True if `obj` is in the Id cache, False otherwise.
        """
        return obj in self._id_cache

    def get_all_models(self) -> Set[BaseModel]:
        """
        Returns a set of all models in cache.

        :return: The set containing all models.
        """
        return set(self._id_cache.copy())

    def _search_key_for_model(
        self, obj: BaseModel, deep_search: bool = True
    ) -> Optional[KeyCacheKey]:

        key_hash = self._get_type_key_hash(obj)
        has_empty_pk = self._has_empty_pk(key_hash)

        # optimize by accessing the cache directly and checking if the model registered
        # with the given key is the actual object we are looking for
        if not has_empty_pk and self.get_by_primary_key(key_hash) == obj:
            return key_hash

        # if not found, then try to find them directly by iterating over the stored
        # models
        return self._search_iterative(obj) if deep_search else None

    def _search_iterative(self, obj: BaseModel) -> Optional[KeyCacheKey]:
        model_key: Optional[KeyCacheKey] = None

        for model_key, model in self._key_cache.items():
            if model == obj:
                return model_key

        return None

    def _get_type_key_hash(self, obj: BaseModel) -> KeyCacheKey:
        _check_model_type(obj)

        obj_type = obj.__class__
        obj_pk = tuple(obj.primary_keys.values())

        return obj_type, obj_pk

    def _has_empty_pk(self, obj_pk: Optional[KeyCacheKey]):
        return not obj_pk or None in obj_pk[1]

    def _remove_from_cache(
        self, obj: BaseModel, cached_model_key: Optional[KeyCacheKey],
    ):
        self._id_cache.discard(obj)
        if cached_model_key and cached_model_key in self._key_cache:
            del self._key_cache[cached_model_key]


class QueryCache:
    """
    Stores query results based on Query hash.
    """

    _cache: Dict[int, Tuple[BaseModel, ...]]

    def __init__(self):
        self.reset()

    def reset(self):
        """
        Resets the internal cache.
        """
        self._cache = {}

    def register(
        self,
        query: BaseQuery[Model_co],
        results: Tuple[Model_co, ...],
        force: bool = False,
    ) -> bool:
        """
        Registers a query in cache.

        :param query: The query instance.
        :param results: The results from the query.
        :param force: Forces registration of the results from query again if it is
                      already registered in cache. Defaults to False
        :return: True if the results have been registered. False otherwise.
        """
        cached = self.get(query) if not force else None

        if not cached:
            self._cache[hash(query)] = results

        return not cached

    def get(self, query: BaseQuery[Model_co]) -> Optional[Tuple[Model_co, ...]]:
        """
        Returns results stored in cache for a particular query.

        :param query: The query instance.
        :return: The results from the query if stored in cache.
        """
        return self._cache.get(hash(query))



from restio.fields.base import ContainerField, Field, IterableField
from restio.fields.typed import (
    BoolField,
    EnumField,
    FloatField,
    FrozenSetField,
    FrozenSetModelField,
    FrozenType,
    IntField,
    ModelField,
    StrField,
    TupleField,
    TupleModelField,
    UUIDField,
)

__all__ = (
    "Field",
    "ContainerField",
    "IterableField",
    "IntField",
    "StrField",
    "BoolField",
    "FloatField",
    "UUIDField",
    "EnumField",
    "TupleField",
    "FrozenSetField",
    "ModelField",
    "TupleModelField",
    "FrozenSetModelField",
    "FrozenType",
)



import inspect
from enum import Flag, auto
from typing import TYPE_CHECKING, Callable, Generic, Optional, Type, TypeVar, Union

from restio.shared import MODEL_TYPE_REGISTRY

if TYPE_CHECKING:
    from restio.model import BaseModel


Model_co = TypeVar("Model_co", bound="BaseModel", covariant=True)
T_co = TypeVar("T_co", bound=object, covariant=True)
SubT = TypeVar("SubT")


def _check_field_value_type(
    type_: Type, name: str, value: T_co, allow_none: bool = False
):
    if allow_none and value is None:
        return

    if not isinstance(value, type_):
        raise TypeError(f"Value of `{name}` should be of type {type_.__name__}")


class MISSING:
    pass


class FrozenType(Flag):
    """
    Indicates when a particular field should not be modified.

    - NEVER: The field can be modified at any time (read and write).
    - UPDATE: The field should be read-only for a model update.
    - CREATE: The field should be read-only when creating the model.
    - ALWAYS: The field is always read-only.
    """

    NEVER = auto()
    UPDATE = auto()
    CREATE = auto()
    ALWAYS = UPDATE | CREATE  # type: ignore


SetterType = Callable[[Model_co, T_co], T_co]

# Base fields


class Field(Generic[T_co], object):
    """
    Base type for Fields.
    """

    type_: Union[Type[T_co], str]
    name: str
    init: bool
    pk: bool
    allow_none: bool
    depends_on: bool
    frozen: FrozenType
    repr: bool
    type_check: bool

    _default: Optional[T_co]
    _default_factory: Optional[Callable[[], T_co]]
    _setter: Optional[SetterType]

    def __init__(
        self,
        type_: Union[Type[T_co], str],
        *,
        pk: bool,
        allow_none: bool,
        depends_on: bool,
        frozen: FrozenType,
        init: bool = True,
        default: Union[Optional[T_co], Type[MISSING]] = MISSING,
        default_factory: Union[Optional[Callable[[], T_co]], Type[MISSING]] = MISSING,
        setter: Optional[SetterType] = None,
        repr: bool = True,
        type_check: bool = True,
    ):
        if default is MISSING and default_factory is MISSING:
            if allow_none:
                default = None

        if default is not MISSING and default_factory is not MISSING:
            raise ValueError(
                "Default value for field provided for both `default` and"
                " `default_factory`"
            )

        self.type_ = type_
        self._default = default
        self._default_factory = default_factory
        self.pk = pk
        self.init = init
        self.allow_none = allow_none
        self.depends_on = depends_on
        self.frozen = frozen
        self.setter(setter)
        self.repr = repr
        self.type_check = type_check

        # post-assignment validation
        self._validate_string_type(self.type_)

    def __set_name__(self, owner, name: str):
        self.name = name

    def __set__(self, instance: "BaseModel", value: T_co):
        if instance._initialized:
            instance._pre_update(self, value)

        value = self._check_value(instance, value)

        if instance._initialized:
            instance._update(self, value)

        instance.__dict__[self.name] = value

    def __get__(self, instance: "BaseModel", cls=None) -> T_co:
        if instance is None:
            return self

        self._store_default(instance, force=False)

        return instance.__dict__[self.name]

    def __delete__(self, instance: "BaseModel") -> None:
        del instance.__dict__[self.name]

    def _store_default(self, instance: "BaseModel", force=False):
        if self.name not in instance.__dict__ or force:
            default = self._check_value(instance, self.default)
            instance.__dict__[self.name] = default

    def _check_value(self, instance: "BaseModel", value: T_co) -> T_co:
        if self.type_check:
            self._cache_string_field_type(instance)
            _check_field_value_type(
                self.type_,  # type: ignore
                self._field_name(instance),
                value,
                allow_none=self.allow_none,
            )
        return self._setter(instance, value) if self._setter is not None else value

    def _cache_string_field_type(self, instance: "BaseModel"):
        # this is a one-off operation, for the first check
        if not isinstance(self.type_, str):
            return

        self.type_ = self._parse_string_field_type(instance, self.type_)  # type: ignore

    def _parse_string_field_type(
        self, instance: "BaseModel", type_: str
    ) -> Union[Type[T_co], Type[SubT]]:
        if type_ not in MODEL_TYPE_REGISTRY:
            raise TypeError(
                f"Provided type alias `{type_}` for field "
                "`{self._field_name(instance)}` is not valid."
            )

        return MODEL_TYPE_REGISTRY[type_]

    def _validate_string_type(self, type_: Union[Type[T_co], str]):
        if not self.type_check:
            return

        if not isinstance(type_, str):
            return

        if not self.depends_on:
            raise TypeError(
                f"Type string `{type_}` is invalid for non-relational field."
            )

    @property
    def default(self) -> T_co:
        """
        Extracts the default value of the field.

        :raises ValueError: When no default value has been set during initialization.
        :return: The default value.
        """
        if not self.has_default:
            raise ValueError(
                f"Can't initialize field {self.name}: default value not provided."
            )

        return (
            self._default_factory()
            if self._default_factory is not MISSING
            else self._default
        )

    @property
    def has_default(self) -> bool:
        """
        Indicates if the field has a default value set during the initialization.

        :return: True if a default value has been set, False otherwise.
        """
        return self._default is not MISSING or self._default_factory is not MISSING

    def setter(self, method: Optional[SetterType]):
        """
        Defines the setter function `method` for the current field. `method` is only
        triggered when a value is assigned to the field through the descriptor
        protocol.

        :param method: The method to be called for setting the value. Method should
                       accept 2 parameters and must return the value to be assigned
                       to the field. The first parameter will contain the instance
                       from which the setter was called, and the second will contain
                       the value assigned.
        :raises ValueError: When the signature of `method` is incorrect.
        :return: The decorated `method`.
        """
        if method is not None:
            signature = inspect.signature(method).parameters
            if len(signature) != 2:
                raise ValueError(
                    f"The provided setter {method.__name__} should accept exactly 2"
                    " parameters."
                )
        self._setter = method
        return method

    def _field_name(self, instance: "BaseModel") -> str:
        return f"{instance.__class__.__name__}.{self.name}"


class ContainerField(Field[T_co], Generic[T_co, SubT]):
    sub_type: Union[Type[SubT], str]

    def __init__(
        self,
        type_: Union[Type[T_co], str],
        sub_type: Union[Type[SubT], str],
        *,
        pk: bool,
        allow_none: bool,
        depends_on: bool,
        frozen: FrozenType,
        init: bool = True,
        default: Union[Optional[T_co], Type[MISSING]] = MISSING,
        default_factory: Union[Optional[Callable[[], T_co]], Type[MISSING]] = MISSING,
        setter: Optional[SetterType] = None,
        repr: bool = True,
        type_check: bool = True,
    ) -> None:
        super().__init__(
            type_=type_,
            pk=pk,
            init=init,
            default=default,
            default_factory=default_factory,
            allow_none=allow_none,
            depends_on=depends_on,
            frozen=frozen,
            setter=setter,
            repr=repr,
            type_check=type_check,
        )
        self.sub_type = sub_type

        self._validate_string_type(self.sub_type)

    def _check_sub_value(self, instance: "BaseModel", sub_value: SubT):
        if self.type_check:
            self._cache_string_field_type(instance)

            _check_field_value_type(
                self.sub_type,  # type: ignore
                self.name,
                sub_value,
            )

    def _cache_string_field_type(self, instance: "BaseModel"):
        super()._cache_string_field_type(instance)

        if not isinstance(self.sub_type, str):
            return

        self.sub_type = self._parse_string_field_type(  # type: ignore
            instance, self.sub_type
        )


class IterableField(ContainerField[T_co, SubT]):
    def __init__(
        self,
        type_: Union[Type[T_co], str],
        sub_type: Union[Type[SubT], str],
        *,
        allow_none: bool,
        depends_on: bool,
        frozen: FrozenType,
        init: bool = True,
        default: Union[Optional[T_co], Type[MISSING]] = MISSING,
        default_factory: Union[Optional[Callable[[], T_co]], Type[MISSING]] = MISSING,
        setter: Optional[SetterType] = None,
        repr: bool = True,
        type_check: bool = True,
    ) -> None:
        super().__init__(
            type_=type_,
            sub_type=sub_type,
            pk=False,
            init=init,
            default=default,
            default_factory=default_factory,
            allow_none=allow_none,
            depends_on=depends_on,
            frozen=frozen,
            setter=setter,
            repr=repr,
            type_check=type_check,
        )

    def _check_value(self, instance: "BaseModel", value: SubT):
        if self.type_check:
            value = super()._check_value(instance, value)
            for item in value:
                super()._check_sub_value(instance, item)
        return value



from __future__ import annotations

import asyncio
from collections import deque
from typing import (
    AsyncGenerator,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    cast,
)

# As a design decision, the classes in this file are bound to BaseModel
# to facilitate the navigation across different objects by utilizing
# the already built-in functionality get_children() and to allow monitoring
# of model states. If future implementation depends on drawing dependency
# graphs, then this file should be modified to decouple from BaseModel
# implementation. The node objects in this case will have to be Hashable.
from restio.model import BaseModel

Model_co = TypeVar("Model_co", bound=BaseModel, covariant=True)


class Node(Generic[Model_co]):
    """
    Represents a Node in a Tree of a DependencyGraph.

    Each Node instance stores a BaseModel object that represents a model in the
    dependency Tree. `parents` and `children` store references to the nodes immediately
    above and below them in the Tree.
    """

    node_object: Model_co
    parents: Set[Node[Model_co]]
    children: Set[Node[Model_co]]

    def __init__(
        self,
        node_object: Model_co,
        parents: Optional[Set[Node[Model_co]]] = None,
        children: Optional[Set[Node[Model_co]]] = None,
    ):
        self.node_object = node_object
        self.parents = parents if parents else set()
        self.children = children if children else set()

    def get_children(
        self, recursive: bool = False, children: Optional[Set[Node[Model_co]]] = None
    ) -> Set[Node[Model_co]]:
        """
        Returns the child nodes of the current Node.

        :param recursive: Indicates whether all child nodes should be returned.
                          Defaults to False.
        :param children: Contains the nodes that have already been inspected and should
                         be ignored. Used for recursion only.
        :return: Returns all children nodes, including children of children, if
                 `recursive` is True. The operation stops when all leaves are reached.
                 Returns only the first degree children if False.
        """
        return self._get_nodes("children", recursive=recursive, nodes=children)

    def get_parents(
        self, recursive: bool = False, parents: Optional[Set[Node[Model_co]]] = None
    ) -> Set[Node[Model_co]]:
        """
        Returns the parent nodes of the current Node.

        :param recursive: Indicates whether all parent nodes should be returned.
                          Defaults to False.
        :param parents: Contains the nodes that have already been inspected and should
                        be ignored. Used for recursion only.
        :return: Returns all parent nodes, including parents of parents, if `recursive`
                 is True. The operation stops when roots are reached. Returns only the
                 first degree parents if False.
        """
        return self._get_nodes("parents", recursive=recursive, nodes=parents)

    def _get_nodes(
        self,
        nodes_attribute: str,
        recursive: bool = False,
        nodes: Optional[Set[Node[Model_co]]] = None,
    ) -> Set[Node]:
        dependent_nodes = getattr(self, nodes_attribute, [])

        if not recursive:
            return dependent_nodes.copy()

        if not nodes:
            nodes = set()

        for node in dependent_nodes:
            if node not in nodes:
                nodes.add(node)
                nodes = nodes.union(node._get_nodes(nodes_attribute, recursive, nodes))

        return nodes

    def __hash__(self):
        return self.node_object.__hash__()

    def __eq__(self, other):
        if not isinstance(other, Node):
            return False

        if not self.node_object or not other.node_object:
            return False

        return self.node_object.__hash__() == other.node_object.__hash__()


GetRelativesCallable = Callable[..., Set[Node[Model_co]]]
NavigationDirection = Tuple[GetRelativesCallable, GetRelativesCallable]


class NavigationType:
    """
    Indicates how the Tree navigation should be done.

    - ROOTS_TO_LEAVES: Starts at the roots of the tree and moves towards the leaves.
    - LEAVES_TO_ROOTS: Starts at the leaves of the tree and moves towards the roots.
    """

    ROOTS_TO_LEAVES: NavigationDirection = (
        Node.get_parents,
        Node.get_children,
    )
    LEAVES_TO_ROOTS: NavigationDirection = (
        Node.get_children,
        Node.get_parents,
    )


class Tree(Generic[Model_co]):
    """
    Represents a Tree in a DependencyGraph.

    Each Tree stores a set of Nodes that have at least one degree of relationship with
    each other. The Tree and can be traversed with callback based tasks using the
    method `process`.
    """

    nodes: Set[Node[Model_co]]
    _canceled: bool
    _processing: bool

    def __init__(self, nodes: Set[Node[Model_co]]):
        self.nodes = nodes
        self._canceled = False
        self._processing = False

    async def navigate(
        self,
        nodes: Set[Node[Model_co]],
        direction: NavigationDirection,
        processed_nodes: asyncio.Queue[Node[Model_co]],
    ) -> AsyncGenerator[Node, bool]:
        """
        Traverses the dependency Tree based on already processed nodes. The caller
        should maintain the queue `processed_nodes` with the nodes that have been
        processed on the past iteration. This generator will yield a Node instance when
        possible, otherwise it will hang until extra nodes processed.

        The Tree traversal order will depend on the `direction` specified (either from
        roots to leaves or leaves to roots) and the order in which nodes are processed
        by the caller.

        :param nodes: The set of Node instances to be processed.
        :param direction: The direction of navigation, either from LEAVES_TO_ROOTS
                          (navigates upwards in the tree) or ROOTS_TO_LEAVES (navigates
                          downwards in the tree).
        :param processed_nodes: The queue containing the nodes that have just been
                                processed by the caller.
        :raises TypeError: If the direction is invalid.
        :yield: The next node in the navigation.
        """

        if direction == NavigationType.LEAVES_TO_ROOTS:
            entrypoint = self.get_leafs()
        elif direction == NavigationType.ROOTS_TO_LEAVES:
            entrypoint = self.get_roots()
        else:
            raise TypeError("The provided argument `direction` is invalid.")

        from_direction, to_direction = direction
        next_nodes = deque(n for n in entrypoint if n in nodes)

        processed: Optional[Node] = None

        while nodes or next_nodes:
            if next_nodes:
                yield next_nodes.pop()
                continue

            processed = await processed_nodes.get()

            if isinstance(processed, Node) and processed not in nodes:
                return

            nodes.remove(processed)

            nodes_to_direction = to_direction(processed) if processed else set()
            for node_to in nodes_to_direction:
                nodes_from_direction = from_direction(node_to)
                if not nodes_from_direction.intersection(nodes):
                    next_nodes.appendleft(node_to)

    def get_roots(self) -> Set[Node[Model_co]]:
        """
        Returns all roots of the Tree.

        :return: Set containing Node instances.
        """
        return self._get_tree_roots(self.nodes)

    def get_leafs(self) -> Set[Node[Model_co]]:
        """
        Returns all leaves of the Tree.

        :return: Set containing Node instances.
        """
        return self._get_tree_leafs(self.nodes)

    def get_nodes(self) -> Set[Node[Model_co]]:
        """
        Returns a copy of all nodes in the Tree.

        :return: Set with Node instances.
        """
        return self.nodes.copy()

    def cancel(self):
        """
        Cancels the processing when `process` is active. Currently running tasks will
        be finalized normally, and new tasks will not be scheduled.
        """
        if self._processing:
            self._canceled = True

    @staticmethod
    def _get_tree_roots(tree_nodes: Set[Node[Model_co]]) -> Set[Node[Model_co]]:
        return set(filter(lambda x: not x.parents, tree_nodes))

    @staticmethod
    def _get_tree_leafs(tree_nodes: Set[Node[Model_co]]) -> Set[Node[Model_co]]:
        return set(filter(lambda x: not x.children, tree_nodes))


class DependencyGraph(Generic[Model_co]):
    """
    Represents dependency graph made of a combination of Tree instances.

    The DependencyGraph stores a list of Tree instances that contain all the Nodes in
    the graph. This module is also responsible to instantiate all Trees and Nodes given
    a set of objects of type BaseModel.
    """

    trees: List[Tree[Model_co]] = []

    def __init__(self, trees: List[Tree[Model_co]]):
        self.trees = trees

    @classmethod
    def generate_from_objects(cls, objects: Set[Model_co]) -> DependencyGraph:
        """
        Generates a DependencyGraph instance based on the set of BaseModel instances
        `objects`.

        :param objects: The set of BaseModel instances.
        :return: The DependencyGraph instance.
        """
        nodes: Set[Node] = cls._get_connected_nodes(objects)
        return cls.generate_from_nodes(nodes)

    @classmethod
    def generate_from_nodes(cls, nodes: Set[Node[Model_co]]) -> DependencyGraph:
        """
        Generates a DependencyGraph instance based on the set of Node instances `nodes`.

        :param nodes: The set of Node instances
        :return: The DependencyGraph instance.
        """
        roots: Set[Node] = Tree._get_tree_roots(nodes)
        roots_children: Dict[Node, Set[Node]] = {
            root: root.get_children(True) for root in roots
        }

        trees: List[Tree] = []

        # generates trees with intersections
        while roots:
            root = roots.pop()
            intersecting_items = {root}.union(roots_children[root])
            intersecting_roots = {root}
            for next_root in roots:
                next_root_children = {next_root}.union(roots_children[next_root])
                if intersecting_items.intersection(next_root_children):
                    intersecting_items = intersecting_items.union(next_root_children)
                    intersecting_roots.add(next_root)

            trees.append(Tree(intersecting_items))
            roots = roots - intersecting_roots

        return cls(trees)

    @staticmethod
    def _get_connected_nodes(objects: Set[Model_co]) -> Set[Node]:
        nodes: Dict[str, Node] = {}

        # creates nodes
        for node_object in objects:
            add_node = Node(node_object)
            nodes[str(node_object.__hash__())] = add_node

        # connects nodes
        for node in nodes.values():
            for child in node.node_object.get_children(recursive=False):
                child_node = nodes.get(str(child.__hash__()), None)
                if not child_node:
                    continue

                node.children.add(child_node)
                child_node.parents.add(node)

        # check for circular dependency
        for node in nodes.values():
            all_children = node.get_children(recursive=True)
            if all_children.intersection(node.get_parents(recursive=False)):
                raise RuntimeError("Circular dependency detected")

        return cast(Set[Node], nodes.values())



import enum
import uuid
from typing import Callable, FrozenSet, Generic, Optional, Tuple, Type, TypeVar, Union

from restio.fields.base import (
    MISSING,
    Field,
    FrozenType,
    IterableField,
    Model_co,
    SetterType,
    SubT,
)

# Constructors are mostly copy-pasted in this file so autocompletion support is more
# user friendly in any IDE


class IntField(Field[int]):
    def __init__(
        self,
        *,
        pk: bool = False,
        init: bool = True,
        default: Union[Optional[int], Type[MISSING]] = MISSING,
        default_factory: Union[Optional[Callable[[], int]], Type[MISSING]] = MISSING,
        allow_none: bool = False,
        frozen: FrozenType = FrozenType.NEVER,
        setter: Optional[SetterType] = None,
        repr: bool = True,
        type_check: bool = True,
    ) -> None:
        super().__init__(
            type_=int,
            pk=pk,
            init=init,
            default=default,
            default_factory=default_factory,
            allow_none=allow_none,
            depends_on=False,
            frozen=frozen,
            setter=setter,
            repr=repr,
            type_check=type_check,
        )


class StrField(Field[str]):
    def __init__(
        self,
        *,
        pk: bool = False,
        init: bool = True,
        default: Union[Optional[str], Type[MISSING]] = MISSING,
        default_factory: Union[Optional[Callable[[], str]], Type[MISSING]] = MISSING,
        allow_none: bool = False,
        frozen: FrozenType = FrozenType.NEVER,
        setter: Optional[SetterType] = None,
        repr: bool = True,
        type_check: bool = True,
    ) -> None:
        super().__init__(
            type_=str,
            pk=pk,
            init=init,
            default=default,
            default_factory=default_factory,
            allow_none=allow_none,
            depends_on=False,
            frozen=frozen,
            setter=setter,
            repr=repr,
            type_check=type_check,
        )


class BoolField(Field[bool]):
    def __init__(
        self,
        *,
        pk: bool = False,
        init: bool = True,
        default: Union[Optional[bool], Type[MISSING]] = MISSING,
        default_factory: Union[Optional[Callable[[], bool]], Type[MISSING]] = MISSING,
        allow_none: bool = False,
        frozen: FrozenType = FrozenType.NEVER,
        setter: Optional[SetterType] = None,
        repr: bool = True,
        type_check: bool = True,
    ) -> None:
        super().__init__(
            type_=bool,
            pk=pk,
            init=init,
            default=default,
            default_factory=default_factory,
            allow_none=allow_none,
            depends_on=False,
            frozen=frozen,
            setter=setter,
            repr=repr,
            type_check=type_check,
        )


class FloatField(Field[float]):
    def __init__(
        self,
        *,
        pk: bool = False,
        init: bool = True,
        default: Union[Optional[float], Type[MISSING]] = MISSING,
        default_factory: Union[Optional[Callable[[], float]], Type[MISSING]] = MISSING,
        allow_none: bool = False,
        frozen: FrozenType = FrozenType.NEVER,
        setter: Optional[SetterType] = None,
        repr: bool = True,
        type_check: bool = True,
    ) -> None:
        super().__init__(
            type_=float,
            pk=pk,
            init=init,
            default=default,
            default_factory=default_factory,
            allow_none=allow_none,
            depends_on=False,
            frozen=frozen,
            setter=setter,
            repr=repr,
            type_check=type_check,
        )


class UUIDField(Field[uuid.UUID]):
    def __init__(
        self,
        *,
        pk: bool = False,
        init: bool = True,
        default: Union[Optional[uuid.UUID], Type[MISSING]] = MISSING,
        default_factory: Union[
            Optional[Callable[[], uuid.UUID]], Type[MISSING]
        ] = MISSING,
        allow_none: bool = False,
        frozen: FrozenType = FrozenType.NEVER,
        setter: Optional[SetterType] = None,
        repr: bool = True,
        type_check: bool = True,
    ) -> None:
        super().__init__(
            type_=uuid.UUID,
            pk=pk,
            init=init,
            default=default,
            default_factory=default_factory,
            allow_none=allow_none,
            depends_on=False,
            frozen=frozen,
            setter=setter,
            repr=repr,
            type_check=type_check,
        )


EnumType = TypeVar("EnumType", bound=enum.Enum)


class EnumField(Field[EnumType], Generic[EnumType]):
    def __init__(
        self,
        type_: Type[EnumType],
        *,
        pk: bool = False,
        init: bool = True,
        default: Union[Optional[EnumType], Type[MISSING]] = MISSING,
        default_factory: Union[
            Optional[Callable[[], EnumType]], Type[MISSING]
        ] = MISSING,
        allow_none: bool = False,
        frozen: FrozenType = FrozenType.NEVER,
        setter: Optional[SetterType] = None,
        repr: bool = True,
        type_check: bool = True,
    ) -> None:
        if not issubclass(type_, enum.Enum):
            raise TypeError("FieldEnum provided type must be a subclass of enum.Enum.")

        super().__init__(
            type_=type_,
            pk=pk,
            init=init,
            default=default,
            default_factory=default_factory,
            allow_none=allow_none,
            depends_on=False,
            frozen=frozen,
            setter=setter,
            repr=repr,
            type_check=type_check,
        )


class TupleField(IterableField[Tuple[SubT, ...], SubT]):
    def __init__(
        self,
        sub_type: Type[SubT],
        *,
        init: bool = True,
        depends_on: bool = False,
        default: Union[Optional[Tuple[SubT, ...]], Type[MISSING]] = MISSING,
        default_factory: Union[
            Optional[Callable[[], Tuple[SubT, ...]]], Type[MISSING]
        ] = MISSING,
        frozen: FrozenType = FrozenType.NEVER,
        setter: Optional[SetterType] = None,
        repr: bool = True,
        type_check: bool = True,
    ) -> None:
        super().__init__(
            type_=tuple,
            sub_type=sub_type,
            init=init,
            default=default,
            default_factory=default_factory,
            allow_none=False,
            depends_on=depends_on,
            frozen=frozen,
            setter=setter,
            repr=repr,
            type_check=type_check,
        )


class FrozenSetField(IterableField[FrozenSet[SubT], SubT]):
    def __init__(
        self,
        sub_type: Type[SubT],
        *,
        init: bool = True,
        depends_on: bool = False,
        default: Union[Optional[FrozenSet[SubT]], Type[MISSING]] = MISSING,
        default_factory: Union[
            Optional[Callable[[], FrozenSet[SubT]]], Type[MISSING]
        ] = MISSING,
        frozen: FrozenType = FrozenType.NEVER,
        setter: Optional[SetterType] = None,
        repr: bool = True,
        type_check: bool = True,
    ) -> None:
        super().__init__(
            type_=frozenset,
            sub_type=sub_type,
            init=init,
            default=default,
            default_factory=default_factory,
            allow_none=False,
            depends_on=depends_on,
            frozen=frozen,
            setter=setter,
            repr=repr,
            type_check=type_check,
        )


class ModelField(Field[Model_co]):
    def __init__(
        self,
        model_type: Union[Type[Model_co], str],
        *,
        init: bool = True,
        default: Union[Optional[Model_co], Type[MISSING]] = MISSING,
        default_factory: Union[
            Optional[Callable[[], Model_co]], Type[MISSING]
        ] = MISSING,
        allow_none: bool = True,
        depends_on: bool = True,
        frozen: FrozenType = FrozenType.NEVER,
        setter: Optional[SetterType] = None,
        repr: bool = True,
        type_check: bool = True,
    ) -> None:
        super().__init__(
            type_=model_type,
            pk=False,
            init=init,
            default=default,
            default_factory=default_factory,
            allow_none=allow_none,
            depends_on=depends_on,
            frozen=frozen,
            setter=setter,
            repr=repr,
            type_check=type_check,
        )


class TupleModelField(TupleField[Model_co]):
    def __init__(
        self,
        model_type: Union[Type[Model_co], str],
        *,
        init: bool = True,
        default: Union[Tuple[Model_co, ...], Type[MISSING]] = MISSING,
        default_factory: Union[
            Optional[Callable[[], Tuple[Model_co, ...]]], Type[MISSING]
        ] = MISSING,
        depends_on: bool = True,
        frozen: FrozenType = FrozenType.NEVER,
        setter: Optional[SetterType] = None,
        repr: bool = True,
        type_check: bool = True,
    ) -> None:
        super().__init__(
            sub_type=model_type,  # type: ignore
            init=init,
            default=default,
            default_factory=default_factory,
            depends_on=depends_on,
            frozen=frozen,
            setter=setter,
            repr=repr,
            type_check=type_check,
        )


class FrozenSetModelField(FrozenSetField[Model_co]):
    def __init__(
        self,
        model_type: Union[Type[Model_co], str],
        *,
        init: bool = True,
        default: Union[Optional[FrozenSet[Model_co]], Type[MISSING]] = MISSING,
        default_factory: Union[
            Optional[Callable[[], FrozenSet[Model_co]]], Type[MISSING]
        ] = MISSING,
        depends_on: bool = True,
        frozen: FrozenType = FrozenType.NEVER,
        setter: Optional[SetterType] = None,
        repr: bool = True,
        type_check: bool = True,
    ) -> None:
        super().__init__(
            sub_type=model_type,  # type: ignore
            init=init,
            default=default,
            default_factory=default_factory,
            depends_on=depends_on,
            frozen=frozen,
            setter=setter,
            repr=repr,
            type_check=type_check,
        )



from __future__ import annotations

from collections.abc import Iterable
from reprlib import Repr
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Set, Tuple, Type
from uuid import UUID, uuid4

from restio.event import EventListener
from restio.fields.base import Field, T_co
from restio.shared import (
    CURRENT_SESSION,
    MODEL_INSTANTIATED_EVENT,
    MODEL_PRE_UPDATE_EVENT,
    MODEL_TYPE_REGISTRY,
    MODEL_UPDATE_EVENT,
)
from restio.state import ModelState

if TYPE_CHECKING:
    from restio.session import Session


def _check_model_type(obj: Optional[BaseModel]):
    if not isinstance(obj, BaseModel):
        raise TypeError("The provided object is not of type BaseModel.")


class ModelMeta:
    __slots__ = ("init", "init_ignore_extra", "repr", "fields", "primary_keys", "alias")

    init: bool
    init_ignore_extra: bool
    repr: bool
    fields: Dict[str, Field]
    primary_keys: Dict[str, Field]
    alias: Optional[str]

    def __init__(self):
        self.init = True
        self.init_ignore_extra = True
        self.repr = True
        self.fields = dict()
        self.primary_keys = dict()
        self.alias = None


# Meta attributes that don't get inherited from parent classes
__MODEL_META_NOT_INHERITED__ = ("alias",)
# Read-only meta attributes, can't be modified by model class
__MODEL_META_READONLY__ = ("fields", "primary_keys")


class BaseModelMeta(type):
    __slots__ = ()

    """
    BaseModel metaclass. Responsible to internally cache the data schema in a BaseModel
    subclass by identifying fields and primary keys.
    """

    def __new__(cls, name: str, bases: Tuple[Type, ...], dct: Dict[str, Any]):
        # internal fields not initialized in BaseModel
        dct["_internal_id"] = None
        dct["_hash"] = None
        dct["_listener"] = None
        dct["_persistent_values"] = None

        # prepares metadata for the model type
        meta = ModelMeta()
        dct["_meta"] = meta

        def _update_meta(
            _meta: Optional[ModelMeta],
            extend: bool,
            not_inherited: Tuple[str, ...] = tuple(),
        ):
            if not _meta:
                return

            propagate_meta = (
                set(meta.__slots__) - set(__MODEL_META_READONLY__) - set(not_inherited)
            )

            for meta_attribute in propagate_meta:
                if not hasattr(_meta, meta_attribute):
                    continue

                setattr(meta, meta_attribute, getattr(_meta, meta_attribute))

            # excluded meta, needs to be propagated manually
            if extend:
                meta.fields.update(_meta.fields)
                meta.primary_keys.update(_meta.primary_keys)

        base: Type[BaseModel]
        for base in bases:
            if not hasattr(base, "_meta"):
                continue

            _update_meta(base._meta, True, __MODEL_META_NOT_INHERITED__)

        _update_meta(dct.get("Meta", None), False)

        # process class fields
        for field_name, field_value in dct.items():
            if not isinstance(field_value, Field):
                continue

            meta.fields[field_name] = field_value
            if field_value.pk:
                meta.primary_keys[field_name] = field_value

        # set alias name to class name when None
        name_alias = meta.alias or name

        # validate if the alias is not duplicate
        # the caveat here is that two classes with the same name in two
        # different files will have a name collision and fail initializing
        if name_alias in MODEL_TYPE_REGISTRY:
            raise ValueError(
                f"Model alias `{name_alias}` is already used by another class."
            )

        cls_object = super().__new__(cls, name, bases, dct)

        # set the model alias to the model type
        if name_alias != "BaseModel":
            MODEL_TYPE_REGISTRY[name_alias] = cls_object

        return cls_object

    def __call__(self, *args, **kwargs):
        instance: BaseModel = super().__call__(*args, **kwargs)

        # stores the default after the constructor, if nothing has been set yet
        # this is implemented here so that this is always called, regardless of the
        # models with custom constructors calling or not super().__init__()
        for field in instance._meta.fields.values():
            field._store_default(instance, force=False)

        instance._internal_id = uuid4()
        instance._hash = hash((instance.__class__, str(instance._internal_id)))
        instance._persistent_values = {}
        instance._listener = EventListener()
        instance._initialized = True

        session = CURRENT_SESSION.get()
        if session:
            session._listener.dispatch(MODEL_INSTANTIATED_EVENT, instance)

        return instance


_repr_obj: Repr = Repr()
_repr_obj.maxother = 200


class BaseModel(metaclass=BaseModelMeta):
    """
    A representation of a remote object model.

    BaseModel is an abstract class that should be extended to represent models incoming
    from or outgoing to a remote REST API.

    Models can exist independently from Sessions but contain an internal state that
    indicates the status of the model within the current context. The Sessions are
    responsible to control this state. Also, each model contains a set of control
    attributes that indicate which fields are watched by restio internals. By default,
    all Field descriptors in the model will become field attributes. Fields declared
    with pk=True will be used by restio to optimize the caching of the models in a
    Session.

    Models that change over time will contain an internal dictionary with the latest
    know persistent value of each field. This is done to guarantee fast rollback of the
    values when the Session is invalid, and to also indicate which values might have
    changed within the session scope. If a field is modified directly, the model will
    intercept the change and save the older value into the persistent dictionary until
    `_persist` is called. During a `_rollback` call, however, the stored values are
    re-assigned to their original attributes. Each attribute change will also dispatch
    an update event so that the session is aware of changes and manages the model's
    internal state accordingly. The persistent dictionary (through the helper method
    `is_field_modified`) can also be used by DAO's to verify which values where updated
    prior to sending a request through the REST API, thus allowing for proper
    optimization and minimizing chances of conflicting changes on the remote object.

    All models automatically generate a random internal UUID when created. This UUID is
    used internally for comparison purposes, and externally as an identity. Although
    this attribute is not explicitly set as private, it should never be modified.
    """

    # these are all initialized by the metaclass
    _meta: ModelMeta

    __state: ModelState = ModelState.UNBOUND
    __primary_keys: Optional[Dict[str, Any]] = None
    _initialized: bool = False

    _internal_id: UUID
    _hash: int
    _persistent_values: Dict[str, Any]
    _listener: EventListener

    def __init__(self, **kwargs: T_co):
        """
        Instantiates the model by matching `kwargs` parameters to field names.
        Behavior is disabled when init=False in the model Meta class.

        :param kwargs: The dictionary of keyword arguments matching the field names of
                       the model class.
        :raises ValueError: When invalid arguments are provided.
        """
        meta = self._meta

        if not meta.init:
            return

        for arg_name, value in kwargs.items():
            field_object = meta.fields.get(arg_name, None)

            if not field_object:
                if not meta.init_ignore_extra:
                    raise ValueError(
                        "Invalid argument provided to constructor of"
                        f" `{self.__class__.__name__}`: {arg_name}"
                    )
                continue  # pragma: no cover

            if not field_object.init:
                if not meta.init_ignore_extra:
                    raise ValueError(f"Attribute `{arg_name}` cannot be initialized.")
                continue  # pragma: no cover

            field_object.__set__(self, value)

    @property
    def _state(self) -> ModelState:
        """
        Returns the state of the current model.

        :return: The ModelState representation.
        """
        return self.__state

    @_state.setter
    def _state(self, state: ModelState):
        self.__state = state

    @property
    def primary_keys(self) -> Dict[str, T_co]:
        """
        Returns a dictionary containing all primary keys. The keys will be
        ordered in the same order as they are declared in the model type,
        also following the order in which they appear in class inheritance.

        This property is optimized to minimize the number of iterations done
        in the model instance by internalizing a cache with the latest retrieved
        primary keys. This cache is reset for every modification of a primary
        key and recovered during the next call to the property.

        :return: The ordered tuple of values.
        """
        if self.__primary_keys is None:
            self.__primary_keys = self._load_primary_keys()

        return self.__primary_keys

    def _load_primary_keys(self) -> Dict[str, T_co]:
        """
        Returns a dictionary containing the primary key fields (keys) and their
        current values in the model (values). This operation will inspect the
        instance and collect all current values on-spot.

        :return: Dictionary of primary keys values.
        """
        return {key: getattr(self, key) for key in self._meta.primary_keys}

    def _reset_primary_keys(self):
        """
        Resets the internal cache of primary keys for the instance.
        """
        self.__primary_keys = None

    def get_children(
        self,
        recursive: bool = False,
        children: Optional[Set[BaseModel]] = None,
        top_level: Optional[BaseModel] = None,
    ) -> Set[BaseModel]:
        """
        Returns the list of all children of the current model. This algorithm checks in
        runtime for all objects refered by the instance and that are part of fields
        marked with depends_on=True. When `recursive` is True, then the algorithm will
        recursively search through all children.

        `children` and `top_level` are control variables that indicate which models
        have already been inspected by this function, in order to avoid infinite
        recursion if any circular dependency exists. In most cases, they should be left
        empty.

        :param recursive: If True, recursively searches for children. Returns only
                          first degree relationships otherwise. Defaults to False.
        :param children: List of existing models already inspected.
        :param top_level: The top-level model from where inspection started.
        :return: The list of children.
        """

        if children is None:
            children = set()

        if top_level:
            if self == top_level:
                return children

            children.add(self)
        else:
            top_level = self

        for value in self.dependency_fields.values():

            def check(child: Optional[BaseModel]):
                # this can happen when the field allows none
                if not child or child in children:  # type: ignore
                    return

                if recursive:
                    child.get_children(recursive, children, top_level)
                else:
                    children.add(child)

            # iterables are only supported if the values are not iterables - there is
            # no recursiveness
            if isinstance(value, Iterable):
                value: Iterable[Any]
                for item in value:
                    check(item)
            else:
                check(value)

        return children

    @property
    def fields(self) -> Dict[str, Any]:
        """
        Returns the values of each field in the model instance.

        :return: A dict with keys containing the string names of the fields,
                 and values containing the value of the corresponding field.
        """
        return {k: getattr(self, k) for k in self._filter_fields(lambda v: True)}

    @property
    def dependency_fields(self) -> Dict[str, Any]:
        """
        Returns the values of each field that have relationship with other models.

        :return: The dictionary of fields and their values
        """
        return {
            k: getattr(self, k) for k in self._filter_fields(lambda v: v.depends_on)
        }

    def is_field_modified(self, field_name: str) -> bool:
        """
        Indicates of field with name `field_name` has been modified.

        :param field_name: The name of the field.
        :raises ValueError: When the field name does not exist.
        :return: True if field is modified, False otherwise.
        """
        if field_name not in self._meta.fields:
            raise ValueError(
                f"Field `{field_name}` does not exist in model"
                " `{self.__class__.__name__}`."
            )

        return field_name in self._persistent_values

    def _filter_fields(self, filt: Callable[[Field], bool]):
        return {k: v for k, v in self._meta.fields.items() if filt(v)}

    def _rollback(self):
        """
        Restore the persistent values in the model to their original attributes.
        """
        for attr, value in list(self._persistent_values.items()):
            setattr(self, attr, value)

        self._persist()

    def _persist(self):
        """
        Persists the current attribute values by emptying the internal persistent
        dictionary. Once this is called, it is not possible to rollback to the old
        values anymore. It is recommended that this method should only be called by the
        party that persisted the values on the remote server.
        """
        self._persistent_values = {}

    def _pre_update(self, field: Field[T_co], value: T_co):
        self._listener.dispatch(MODEL_PRE_UPDATE_EVENT, self, field, value)

    def _update(self, field: Field[T_co], value: T_co):
        if field.pk:
            self._reset_primary_keys()

        self._listener.dispatch(MODEL_UPDATE_EVENT, self, field, value)

    def _update_persistent_values(self, field: Field[T_co], value: T_co):
        name: str = field.name
        if name in self._persistent_values:
            if value == self._persistent_values[name]:
                del self._persistent_values[name]
        else:
            mutable_fields = self.fields
            if value != mutable_fields[name]:
                self._persistent_values[name] = mutable_fields[name]

    def __eq__(self, other: BaseModel) -> bool:
        return isinstance(other, self.__class__) and self._hash == other._hash

    def __repr__(self) -> str:
        if not self._meta.repr:
            return super().__repr__()

        def get_field_repr(field: str):
            value = getattr(self, field)
            return f"{field}={_repr_obj.repr(value)}"

        repr_args: List[str] = [
            get_field_repr(n) for n in self._filter_fields(lambda x: x.repr)
        ]
        return f"{self.__class__.__name__}({', '.join(repr_args)})"

    def __hash__(self) -> int:
        return self._hash



from __future__ import annotations

import asyncio
import time
from functools import wraps
from typing import TYPE_CHECKING, Awaitable, Callable, Generic, Optional, Type, TypeVar

from restio.graph import Node
from restio.model import BaseModel

NOT_IMPLEMENTED_DAO_FUNC_ATTR = "dao_method_not_implemented"

if TYPE_CHECKING:
    from restio.session import Session


FunctionType = Callable[..., Awaitable]


def not_implemented_method(function: FunctionType) -> FunctionType:
    @wraps(function)
    def wrapper(self, obj):
        raise NotImplementedError(f"Method {function.__name__} not implemented.")

    setattr(wrapper, NOT_IMPLEMENTED_DAO_FUNC_ATTR, True)
    return wrapper


def check_dao_implemented_method(function: FunctionType):
    if hasattr(function, NOT_IMPLEMENTED_DAO_FUNC_ATTR) and getattr(
        function, NOT_IMPLEMENTED_DAO_FUNC_ATTR
    ):
        raise NotImplementedError(
            f"Function `{function.__name__}` of DAO"
            f" {function.__self__.__class__.__name__} not implemented."
        )


Model_co = TypeVar("Model_co", bound=BaseModel, covariant=True)


class BaseDAO(Generic[Model_co]):
    """
    Base abstract class for Data Access Objects (DAO).

    The subclasses of BaseDAO in the restio framework represent the core of the data
    access layer to a particular model in a remote REST API. Sessions use the DAOs
    to perform CRUD operations on the remote server using the async methods `get`,
    `add`, `update` and `remove`.

    Each DAO instance should be registered to a Session instance and associated to a
    model type through its constructor. The model type is stored internally and used by
    the Session to identify the DAO that is responsible for a particular model. When
    bound to a Session instance, the DAO instance receives a reference to the Session,
    which is always available internally.

    Differently from a regular abstract class, the methods of BaseDAO subclasses don't
    need to be overwritten unless the Session needs to make use of them. In that case,
    not implementing the method will cause the Session to fail in runtime for
    not-implemented `add`, `remove` or `update`. Methods that fail during a Session
    commit might raise an Exception, which will be picked up by the Session. Models can
    be modified in the methods of the DAO, and the values will be persisted to the
    model's cache accordingly if the operation is successful (no exception thrown).

    It is recommended that each `get`, `add`, `remove` and `update` method is
    overwritten in the subclass if the operation is permitted by the REST API.
    """

    _model_type: Type[Model_co]
    _session: Optional["Session"]

    def __init__(self, model_type: Optional[Type[Model_co]] = None) -> None:
        if model_type:
            self._model_type = model_type

    @property
    def session(self) -> "Session":
        """
        Returns the Session instance to which the DAO instance is bound.

        :raises RuntimeError: When the DAO instance is not bound to any Session
                              instance.
        :return: The Session instance.
        """
        if not self._session:
            raise RuntimeError("This instance of DAO is not bound to any Session.")
        return self._session

    @session.setter
    def session(self, session: "Session"):
        self._session = session

    @not_implemented_method
    async def get(self, **keys) -> Model_co:  # type: ignore
        """
        Retrieves a model from the remote server.

        # noqa: DAR202 return

        :param keys: The keyword arguments representing the primary keys of the model
                     to be retrieved.
        :return: The model retrieved from the remote server.
        """
        pass

    @not_implemented_method
    async def add(self, obj: Model_co):
        """
        Creates a model in the remote server.

        :param obj: The model instance to be created.
        """
        pass

    @not_implemented_method
    async def remove(self, obj: Model_co):
        """
        Removes a model from the remote server.

        :param obj: The model instance to be removed.
        """
        pass

    @not_implemented_method
    async def update(self, obj: Model_co):
        """
        Updates a model in the remote server.

        :param obj: The model to be updated.
        """
        pass


DAOTaskCallable = Callable[[Model_co], Awaitable[None]]


class DAOTask(Generic[Model_co]):
    """
    Wrapper object that, when awaited, contains a asyncio.Task ran by a a DAO method
    during a Session commit.

    When a DAOTask instance is awaited for the first time, it triggers `run_task()`
    and returns the result from the undelying task wrapping `func`. Awaiting the same
    instance multiple times will not retrigger the task, but instead will return the
    value from the first call. Alternatively, it is also possible to retrieve the
    underlying task directly through the attribute `task`.

    Running a DAOTask will also record the `start_time` and `end_time` of the first
    execution.
    """

    _task: Optional[asyncio.Task[Optional[Model_co]]]
    node: Node[Model_co]
    func: DAOTaskCallable
    start_time: float
    end_time: float

    def __init__(self, node: Node[Model_co], func: DAOTaskCallable):
        self.node = node
        self.func = func
        self.start_time = 0.0
        self.end_time = 0.0
        self._task = None

    def run_task(self) -> asyncio.Task[Optional[Model_co]]:
        """
        Creates and returns an asyncio.Task that runs `func` when called for the first
        time.

        When called multiple times, returns the existing asyncio.Task created during
        the first execution.

        :return: The asyncio.Task instance.
        """
        if self._task:
            return self._task

        self.start_time = time.time()

        task: asyncio.Task = asyncio.create_task(self.func(self.node.node_object))
        task.add_done_callback(self._task_finished)

        self._task = task
        return task

    def __await__(self):
        return (yield from self.run_task().__await__())

    @property
    def task(self) -> asyncio.Task[Optional[Model_co]]:
        """
        Contains the underlying asyncio.Task.

        :raises RuntimeError: When a Task has not yet been triggered.
        :return: The asyncio.Task instance.
        """
        if not self._task:
            raise RuntimeError("DAOTask not started.")
        return self._task

    def _task_finished(self, future):
        self.end_time = time.time()
        self.task.remove_done_callback(self._task_finished)

    @property
    def duration(self) -> float:
        """
        Returns the duration of the execution (in seconds).

        :raises RuntimeError: When the execution has not been finished.
        :return: The duration (in seconds).
        """
        if not self.start_time or not self.end_time:
            raise RuntimeError("Task not finished.")
        return self.end_time - self.start_time

    @property
    def model(self) -> Model_co:
        """
        Returns the BaseModel contained by the `node`.

        :return: The BaseModel instance.
        """
        return self.node.node_object

    def __lt__(self, value: DAOTask):
        return self.end_time < value.end_time



import inspect
from functools import wraps
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Dict,
    Generator,
    Generic,
    Iterable,
    List,
    Optional,
    Tuple,
    TypeVar,
)

from restio.model import BaseModel

if TYPE_CHECKING:
    from restio.session import Session


Model_co = TypeVar("Model_co", bound=BaseModel, covariant=True)

SESSION_KEYWORD = "session"


class BaseQuery(Generic[Model_co]):
    """
    Defines a query object to be processed by a Session.

    Every BaseQuery instance has an internal hash that is generated based on the
    function to be invoked and its parameters. This hash is used by Session query cache
    to store results for queries already executed in the context of the instance of the
    Session.

    Two instances of BaseQuery pointing to the same `function` and with the same `args`
    and `kwargs` provided will resolve on the same hash, and therefore are considered
    to be the same.
    """

    __args: Tuple
    __kwargs: Dict[str, Any]
    __session: Optional["Session"]
    __has_session_argument: bool

    def __init__(
        self, function: Callable[..., Awaitable[Iterable[Model_co]]], *args, **kwargs,
    ):
        self.__function = function
        self.__args = args or tuple()
        self.__kwargs = kwargs or {}
        self.__session = None

        # inspect the function to look for the session keyword
        params = inspect.signature(function).parameters
        session_parameter = params.get(SESSION_KEYWORD, None)
        self.__has_session_argument = bool(session_parameter)

    def __hash__(self):
        hash_list: List[Any] = [self.__function]

        signature = inspect.signature(self.__function)

        # adds the session for binding
        kwargs = self.__kwargs
        if self.__has_session_argument:
            kwargs = kwargs.copy()
            kwargs[SESSION_KEYWORD] = None

        parameters = signature.bind(*self.__args, **kwargs)

        # now ignore the session keyword
        parameters.arguments.pop(SESSION_KEYWORD, None)

        hash_list.extend(list(parameters.arguments.items()))

        return hash(tuple(hash_list))

    def _get_function(self):
        return self.__function

    def __eq__(self, other):
        if other and isinstance(other, BaseQuery):
            return self.__hash__() == other.__hash__()

        return False

    def __call__(self, session: "Session"):
        self.__session = session
        return self

    def __await__(self) -> Generator[Any, Iterable[Model_co], Iterable[Model_co]]:
        kwargs = self.__kwargs

        if self.__has_session_argument:
            kwargs = self.__kwargs.copy()
            session = self.__session or self.__kwargs.get(SESSION_KEYWORD, None)

            if not session:
                raise RuntimeError(
                    f"Session not provided for querying {self.__function.__name__}."
                )
            kwargs[SESSION_KEYWORD] = session

        return (yield from self.__function(*self.__args, **kwargs).__await__())


def query(
    function: Callable[..., Awaitable[Iterable[Model_co]]]
) -> Callable[..., BaseQuery[Model_co]]:
    f"""
    Query decorator.

    This should be used to decorate functions that return query results to be stored by
    a Session. Any decorated functions might contain a parameter `{SESSION_KEYWORD}`
    that will receive the Session instance object. If this parameter is provided when
    building the query, it gets overwritten when the query is executed by a
    Session.

    Queries are useful when performing batch operations on the remote server. The
    Session instance will use the query results and store them internally for
    caching purposes.
    """

    @wraps(function)
    def wrapper(*args, **kwargs) -> BaseQuery[Model_co]:
        return BaseQuery(function, *args, **kwargs)

    return wrapper
